#include "defaultplayercontroller.h"

#include "internal.h"
#include "internal/physics.h"

#include <glm/gtx/vector_angle.hpp>
#include <BulletCollision/CollisionDispatch/btCollisionWorld.h>

namespace kek {

static bool castPlayer(glm::vec3 pos, glm::vec3 delta, const btCollisionObject **outBody, float *outFrac, glm::vec3 *outPoint, glm::vec3 *outNormal) {
	btTransform from;
	from.setIdentity();
	from.setOrigin(Physics::fromGLM(pos));

	btTransform to;
	to.setIdentity();
	to.setOrigin(Physics::fromGLM(pos + delta));

	ClosestNotSelfConvexResultCallback cb = ClosestNotSelfConvexResultCallback(from.getOrigin(), to.getOrigin(), kekData.player->physics->body);
	cb.m_collisionFilterGroup = -1;
	cb.m_collisionFilterMask = -1;
	kekData.physics->world->convexSweepTest((btConvexShape *) kekData.player->physics->body->getCollisionShape(), from, to, cb);
	if(!cb.m_hitCollisionObject) return false;

	if(outBody) *outBody = cb.m_hitCollisionObject;
	if(outFrac) *outFrac = cb.m_closestHitFraction;
	if(outPoint) *outPoint = Physics::toGLM(cb.m_hitPointWorld);
	if(outNormal) *outNormal = Physics::toGLM(cb.m_hitNormalWorld);
	return true;
}

bool DefaultPlayerController::checkPlayerGrounded() {
	glm::vec3 point;
	glm::vec3 normal;
	bool onGround = castPlayer(kekData.player->getPosition(), glm::vec3(0,-groundDistance,0), nullptr, nullptr, &point, &normal);
	// TODO: angle check
	return onGround;
}

DefaultPlayerController::DefaultPlayerController() {

}

DefaultPlayerController::~DefaultPlayerController() {

}

glm::vec3 DefaultPlayerController::move(glm::vec3 movement) {
	glm::vec3 pos = kekData.player->getPosition();
	glm::vec3 totalMovement = glm::vec3(0);

	int count = 0;
	while (count < 3) {
		float frac;
		glm::vec3 normal;
		bool collision = castPlayer(pos + totalMovement, movement, nullptr, &frac, nullptr, &normal);
		if(!collision) {
			totalMovement += movement;
			break;
		}

		normal = glm::normalize(normal);

		totalMovement += movement * frac;

		float angle = 90.0f - glm::angle(normal, glm::normalize(movement)) / M_PI * 90.0f;
		if(angle < minSlideAngle) break;

		totalMovement += normal * 0.001f;

		glm::vec3 remainingProjected = glm::normalize(glm::cross(normal, glm::cross(movement, normal))) * glm::length(movement) * (1 - frac);
		movement = remainingProjected;
		count++;
	}

	kekData.player->translate(totalMovement);
	return totalMovement;
}

void DefaultPlayerController::jump() {
	velocity += glm::vec3(0, jumpVelocity, 0);
}

void DefaultPlayerController::update() {
	bool onGround = checkPlayerGrounded();
	kekData.player->onGround = onGround;

	move(velocity * kekData.lastFrameTime);

	if(!onGround) {
		velocity += gravity * kekData.lastFrameTime;
	}else{
		velocity = glm::vec3(0);
	}
}

}
